
import java.awt.*;
import java.awt.image.*;

public class GBDisplay extends Component implements GameboyObserver {
  Gameboy fGameboy;
  MemoryImageSource mis;
  Image image;
  int pixelScale = 1;
  int joyStick = 0xff;
  boolean fTrace;
  
  public GBDisplay(Gameboy gameboy){
    fGameboy = gameboy;
    fGameboy.setObserver(this);
    setBounds(0,0,fGameboy.visWidth,fGameboy.HEIGHT);
  }
  
 public void refreshScreen(Gameboy gameboy){
  drawScreen();
 }
   
  void drawScreen() {
      Graphics g = getGraphics();
      if (g != null) {
        drawScreen(g);
        g.dispose();
      }
   }
   
  public void paint(Graphics g){
    drawScreen(g);
  }
   
  synchronized void drawScreen(Graphics g) {
     if ( mis == null ) {
       byte[] grey = {(byte)0xff, (byte)0x98, (byte)0x58, (byte)0x00, (byte)0xff, (byte)0x98, (byte)0x58, (byte)0x00, (byte)0xff, (byte)0x98, (byte)0x58, (byte)0x00};
       ColorModel cm = new IndexColorModel(8,grey.length,grey, grey, grey);
       mis = new MemoryImageSource(fGameboy.visWidth,fGameboy.HEIGHT,cm, fGameboy.XBuf, fGameboy.visStartX,fGameboy.WIDTH);
       mis.setAnimated(true);
	    }
     if (image == null) {
      image = createImage(mis);
     }
    else
        mis.newPixels();
    g.drawImage(image,0,0,this);
  }

	public boolean handleEvent( Event e ) {
		switch ( e.id ) {
		case Event.MOUSE_DOWN:
			requestFocus();
			return true;
		case Event.KEY_ACTION:
		case Event.KEY_PRESS:
			return doKey( true, e.key, e.modifiers );
		case Event.KEY_ACTION_RELEASE:
		case Event.KEY_RELEASE:
			return doKey( false, e.key, e.modifiers );
		case Event.GOT_FOCUS:
		case Event.LOST_FOCUS:
			resetKeyboard();
			return true;
		}

	  return false;
  }

	public final boolean doKey( boolean down, int ascii, int mods ) {
		boolean    CAPS = ((mods & Event.CTRL_MASK) != 0);
		boolean    SYMB = ((mods & Event.META_MASK) != 0);
		boolean   SHIFT = ((mods & Event.SHIFT_MASK) != 0);

		// Change control versions of keys to lower case
		if ( (ascii >= 1) && (ascii <= 0x27) && SYMB ) {
			ascii += ('a'-1);
		}
    int mask = 0;
		switch ( ascii ) {
		case 27:  toggleTrace(); break;
		case '1':    doKeys(down, Gameboy.JOY_START);    break;
		case '2':    doKeys(down, Gameboy.JOY_SELECT);    break;
		case '3':    doKeys(down, Gameboy.JOY_A);    break;
		case '4':    doKeys(down, Gameboy.JOY_B);    break;
		case Event.UP:    doKeys(down, Gameboy.JOY_U);    break;
		case Event.DOWN:    doKeys(down, Gameboy.JOY_D);    break;
		case Event.LEFT:    doKeys(down, Gameboy.JOY_L);    break;
		case Event.RIGHT:    doKeys(down, Gameboy.JOY_R);    break;
		default: return false;
    }
    return true;   
  }
  
  void toggleTrace(){
    fTrace = ! fTrace;
    int trace = 0;
    if ( fTrace ) {
      trace = 1;
    }
    fGameboy.Trace(trace);
  }
  void doKeys(boolean down, int mask){
    int oldJoy = joyStick;
    if ( down ) joyStick &= ~mask;
    else joyStick |= mask;
    if ( oldJoy != joyStick ){
      fGameboy.setJoystick(joyStick);
      System.err.println(joyStick);
    }
  }
  
  void resetKeyboard(){
    joyStick = 0xff;
    fGameboy.setJoystick(joyStick);
    System.err.println(joyStick);
  }
    
  
	public Dimension minimumSize() {
		int scale  = 1;
		int border = 0;

		return new Dimension(
			fGameboy.visWidth * scale + border*2,
			fGameboy.HEIGHT * scale + border*2
		);
	}
  
}